@@ -85,15 +85,6 @@ class DHCPResourceUpdate(queue.ResourceUpdate):
 
     def __lt__(self, other):
         if other.obj_type == self.obj_type == 'port':
-            # NOTE(ralonsoh): both resources should have "fixed_ips"
-            # information. That key was added to the deleted ports in this
-            # patch but this code runs in the Neutron API (server). Both the
-            # server and the DHCP agent should be updated.
-            # This check could be removed in Y release.
-            if ('fixed_ips' not in self.resource or
-                    'fixed_ips' not in other.resource):
-                return super().__lt__(other)
-
             self_ips = set(str(fixed_ip['ip_address']) for
                            fixed_ip in self.resource['fixed_ips'])
             other_ips = set(str(fixed_ip['ip_address']) for
@@ -124,8 +115,8 @@ class DhcpAgent(manager.Manager):
         # If 'resync_throttle' is configured more than 'resync_interval' by
         # mistake, raise exception and log with message.
         if self.conf.resync_throttle > self.conf.resync_interval:
-            msg = _("DHCP agent must have resync_throttle <= resync_interval")
-            LOG.exception(msg)
+            LOG.exception("DHCP agent must have resync_throttle <= "
+                          "resync_interval")
             raise exceptions.InvalidConfigurationOption(
                 opt_name='resync_throttle',
                 opt_value=self.conf.resync_throttle)
@@ -246,7 +237,7 @@ class DhcpAgent(manager.Manager):
                 # allocation failure. When the subnet is updated with a new
                 # allocation pool or a port is  deleted to free up an IP, this
                 # will automatically be retried on the notification
-                self.schedule_resync(e, network.id)
+                self.schedule_resync(str(e), network.id)
             if (isinstance(e, oslo_messaging.RemoteError) and
                     e.exc_type == 'NetworkNotFound' or
                     isinstance(e, exceptions.NetworkNotFound)):
@@ -371,7 +362,7 @@ class DhcpAgent(manager.Manager):
                         net = "*"
                     LOG.debug("resync (%(network)s): %(reason)s",
                               {"reason": r, "network": net})
-                self.sync_state(reasons.keys())
+                self.sync_state(list(reasons.keys()))
 
     def periodic_resync(self):
         """Spawn a thread to periodically resync the dhcp state."""
@@ -409,16 +400,14 @@ class DhcpAgent(manager.Manager):
         if not network.admin_state_up:
             return
 
-        for subnet in network.subnets:
-            if subnet.enable_dhcp:
-                if self.call_driver('enable', network):
-                    self.update_isolated_metadata_proxy(network)
-                    self.cache.put(network)
-                    # After enabling dhcp for network, mark all existing
-                    # ports as ready. So that the status of ports which are
-                    # created before enabling dhcp can be updated.
-                    self.dhcp_ready_ports |= {p.id for p in network.ports}
-                break
+        if (any(s for s in network.subnets if s.enable_dhcp) and
+                self.call_driver('enable', network)):
+            self.update_isolated_metadata_proxy(network)
+            self.cache.put(network)
+            # After enabling dhcp for network, mark all existing
+            # ports as ready. So that the status of ports which are
+            # created before enabling dhcp can be updated.
+            self.dhcp_ready_ports |= {p.id for p in network.ports}
 
         self._resize_process_pool()
 
@@ -535,30 +524,9 @@ class DhcpAgent(manager.Manager):
     # Use the update handler for the subnet create event.
     subnet_create_end = subnet_update_end
 
-    def _get_network_lock_id(self, payload):
-        """Determine which lock to hold when servicing an RPC event"""
-        # TODO(alegacy): in a future release this function can be removed and
-        # uses of it can be replaced with payload['network_id'].  It exists
-        # only to satisfy backwards compatibility between older servers and
-        # newer agents.  Once the 'network_id' attribute is guaranteed to be
-        # sent by the server on all *_delete_end events then it can be removed.
-        if 'network_id' in payload:
-            return payload['network_id']
-        elif 'subnet_id' in payload:
-            subnet_id = payload['subnet_id']
-            network = self.cache.get_network_by_subnet_id(subnet_id)
-            return network.id if network else None
-        elif 'port_id' in payload:
-            port_id = payload['port_id']
-            port = self.cache.get_port_by_id(port_id)
-            return port.network_id if port else None
-
     def subnet_delete_end(self, context, payload):
         """Handle the subnet.delete.end notification event."""
-        network_id = self._get_network_lock_id(payload)
-        if not network_id:
-            return
-        update = DHCPResourceUpdate(network_id,
+        update = DHCPResourceUpdate(payload['network_id'],
                                     payload.get('priority', DEFAULT_PRIORITY),
                                     action='_subnet_delete',
                                     resource=payload, obj_type='subnet')
@@ -567,9 +535,6 @@ class DhcpAgent(manager.Manager):
     @_wait_if_syncing
     @log_helpers.log_method_call
     def _subnet_delete(self, payload):
-        network_id = self._get_network_lock_id(payload)
-        if not network_id:
-            return
         subnet_id = payload['subnet_id']
         network = self.cache.get_network_by_subnet_id(subnet_id)
         if not network:
@@ -709,10 +674,7 @@ class DhcpAgent(manager.Manager):
 
     def port_delete_end(self, context, payload):
         """Handle the port.delete.end notification event."""
-        network_id = self._get_network_lock_id(payload)
-        if not network_id:
-            return
-        update = DHCPResourceUpdate(network_id,
+        update = DHCPResourceUpdate(payload['network_id'],
                                     payload.get('priority', DEFAULT_PRIORITY),
                                     action='_port_delete',
                                     resource=payload, obj_type='port')
@@ -721,9 +683,6 @@ class DhcpAgent(manager.Manager):
     @_wait_if_syncing
     @log_helpers.log_method_call
     def _port_delete(self, payload):
-        network_id = self._get_network_lock_id(payload)
-        if not network_id:
-            return
         port_id = payload['port_id']
         port = self.cache.get_port_by_id(port_id)
         self.cache.add_to_deleted_ports(port_id)
@@ -809,7 +768,7 @@ class DhcpAgent(manager.Manager):
                     'Cannot serve metadata on IPv6 because DHCP driver '
                     'does not implement method '
                     'get_metadata_bind_interface(): %s',
-                    self.dhcp_driver_class)
+                    self.dhcp_driver_cls)
 
         metadata_driver.MetadataDriver.spawn_monitored_metadata_proxy(
             self._process_monitor, network.namespace, constants.METADATA_PORT,
