@@ -17,11 +17,13 @@ import abc
 import time
 
 import netaddr
+from neutron_lib.agent.linux import interface
 from neutron_lib import constants
 from neutron_lib import exceptions
 from oslo_log import log as logging
 from oslo_utils import excutils
-from pyroute2.netlink import exceptions as pyroute2_exc
+from pyroute2.netlink import exceptions \
+    as pyroute2_exc  # pylint: disable=no-name-in-module
 
 from neutron.agent.common import ovs_lib
 from neutron.agent.linux import ip_lib
@@ -39,10 +41,8 @@ def _get_veth(name1, name2, namespace2):
             ip_lib.IPDevice(name2, namespace=namespace2))
 
 
-class LinuxInterfaceDriver(object, metaclass=abc.ABCMeta):
-
-    DEV_NAME_LEN = constants.LINUX_DEV_LEN
-    DEV_NAME_PREFIX = constants.TAP_DEVICE_PREFIX
+class LinuxInterfaceDriver(interface.LinuxInterfaceDriver,
+                           metaclass=abc.ABCMeta):
 
     def __init__(self, conf, **kwargs):
         self.conf = conf
@@ -256,19 +256,12 @@ class LinuxInterfaceDriver(object, metaclass=abc.ABCMeta):
                {'dev': dev_name, 'enabled': int(enabled)}]
         ip_lib.sysctl(cmd, namespace=namespace)
 
-    @abc.abstractmethod
-    def plug_new(self, network_id, port_id, device_name, mac_address,
-                 bridge=None, namespace=None, prefix=None, mtu=None,
-                 link_up=True):
-        """Plug in the interface only for new devices that don't exist yet."""
-
     def plug(self, network_id, port_id, device_name, mac_address,
              bridge=None, namespace=None, prefix=None, mtu=None, link_up=True):
         if not ip_lib.device_exists(device_name,
                                     namespace=namespace):
-            self._safe_plug_new(
-                network_id, port_id, device_name, mac_address, bridge,
-                namespace, prefix, mtu, link_up)
+            self.plug_new(network_id, port_id, device_name, mac_address,
+                          bridge, namespace, prefix, mtu, link_up)
         else:
             LOG.info("Device %s already exists", device_name)
             if mtu:
@@ -277,26 +270,6 @@ class LinuxInterfaceDriver(object, metaclass=abc.ABCMeta):
             else:
                 LOG.warning("No MTU configured for port %s", port_id)
 
-    def _safe_plug_new(self, network_id, port_id, device_name, mac_address,
-            bridge=None, namespace=None, prefix=None, mtu=None, link_up=True):
-        try:
-            self.plug_new(
-                network_id, port_id, device_name, mac_address, bridge,
-                namespace, prefix, mtu, link_up)
-        except TypeError:
-            LOG.warning("Interface driver's plug_new() method should now "
-                        "accept additional optional parameter 'link_up'. "
-                        "Usage of plug_new() method which takes from 5 to 9 "
-                        "positional arguments is now deprecated and will not "
-                        "be possible in W release.")
-            self.plug_new(
-                network_id, port_id, device_name, mac_address, bridge,
-                namespace, prefix, mtu)
-
-    @abc.abstractmethod
-    def unplug(self, device_name, bridge=None, namespace=None, prefix=None):
-        """Unplug the interface."""
-
     @property
     def bridged(self):
         """Whether the DHCP port is bridged to the VM TAP interfaces.
@@ -322,20 +295,14 @@ class LinuxInterfaceDriver(object, metaclass=abc.ABCMeta):
 
     def set_link_status(self, device_name, namespace=None, link_up=True):
         ns_dev = ip_lib.IPWrapper(namespace=namespace).device(device_name)
-        try:
-            utils.wait_until_true(ns_dev.exists, timeout=3)
-        except utils.WaitTimeout:
-            LOG.debug('Device %s may have been deleted concurrently',
-                      device_name)
-            return False
-
+        if not ns_dev.exists():
+            LOG.debug("Device %s may concurrently be deleted.", device_name)
+            return
         if link_up:
             ns_dev.link.set_up()
         else:
             ns_dev.link.set_down()
 
-        return True
-
 
 class NullDriver(LinuxInterfaceDriver):
     def plug_new(self, network_id, port_id, device_name, mac_address,
@@ -375,6 +342,48 @@ class OVSInterfaceDriver(LinuxInterfaceDriver):
         ovs = ovs_lib.OVSBridge(bridge)
         ovs.replace_port(device_name, *attrs)
 
+    def _set_device_address(self, device, mac_address):
+        for i in range(9):
+            # workaround for the OVS shy port syndrome. ports sometimes
+            # hide for a bit right after they are first created.
+            # see bug/1618987
+            try:
+                device.link.set_address(mac_address)
+                break
+            except RuntimeError as e:
+                LOG.warning("Got error trying to set mac, retrying: %s",
+                            str(e))
+                time.sleep(1)
+        else:
+            # didn't break, we give it one last shot without catching
+            device.link.set_address(mac_address)
+
+    def _add_device_to_namespace(self, ip_wrapper, device, namespace):
+        namespace_obj = ip_wrapper.ensure_namespace(namespace)
+        for i in range(9):
+            try:
+                namespace_obj.add_device_to_namespace(device)
+                break
+            except ip_lib.NetworkInterfaceNotFound:
+                # NOTE(slaweq): if the exception was NetworkInterfaceNotFound
+                # then lets try again, otherwise lets simply raise it as this
+                # is some different issue than retry tries to workaround
+                LOG.warning("Failed to set interface %s into namespace %s. "
+                            "Interface not found, attempt: %s, retrying.",
+                            device, namespace, i + 1)
+                time.sleep(1)
+            except utils.WaitTimeout:
+                # NOTE(slaweq): if the exception was WaitTimeout then it means
+                # that probably device wasn't found in the desired namespace
+                # for 5 seconds, so lets try again too
+                LOG.warning("Failed to set interface %s into namespace %s. "
+                            "Interface not found in namespace, attempt: %s, "
+                            "retrying.", device, namespace, i + 1)
+                time.sleep(1)
+        else:
+            # didn't break, we give it one last shot without catching
+            namespace_obj.add_device_to_namespace(device)
+
     def plug_new(self, network_id, port_id, device_name, mac_address,
                  bridge=None, namespace=None, prefix=None, mtu=None,
                  link_up=True):
@@ -399,31 +408,25 @@ class OVSInterfaceDriver(LinuxInterfaceDriver):
         internal = not self.conf.ovs_use_veth
         self._ovs_add_port(bridge, tap_name, port_id, mac_address,
                            internal=internal)
-        for i in range(9):
-            # workaround for the OVS shy port syndrome. ports sometimes
-            # hide for a bit right after they are first created.
-            # see bug/1618987
-            try:
-                ns_dev.link.set_address(mac_address)
-                break
-            except RuntimeError as e:
-                LOG.warning("Got error trying to set mac, retrying: %s",
-                            str(e))
-                time.sleep(1)
-        else:
-            # didn't break, we give it one last shot without catching
-            ns_dev.link.set_address(mac_address)
+        try:
+            self._set_device_address(ns_dev, mac_address)
+        except Exception:
+            LOG.warning("Failed to set mac for interface %s", ns_dev)
+            with excutils.save_and_reraise_exception():
+                ovs = ovs_lib.OVSBridge(bridge)
+                ovs.delete_port(tap_name)
 
         # Add an interface created by ovs to the namespace.
         if not self.conf.ovs_use_veth and namespace:
             try:
-                namespace_obj = ip.ensure_namespace(namespace)
-                namespace_obj.add_device_to_namespace(ns_dev)
-            except (pyroute2_exc.NetlinkError, OSError):
+                self._add_device_to_namespace(ip, ns_dev, namespace)
+            except (pyroute2_exc.NetlinkError, OSError, RuntimeError):
                 # To prevent the namespace failure from blasting OVS, the OVS
                 # port creation should be reverted. Possible exceptions:
                 # - NetlinkError in case of duplicated interface
                 # - OSError in case of corrupted namespace
+                # - RuntimeError in case of any issue with interface, like e.g.
+                #   Interface not found
                 LOG.warning("Failed to plug interface %s into bridge %s, "
                             "cleaning up", device_name, bridge)
                 with excutils.save_and_reraise_exception():
