@@ -45,7 +45,7 @@ from sqlalchemy import exc as sql_exc
 from sqlalchemy import orm
 
 from neutron._i18n import _
-from neutron.common import utils as n_utils
+from neutron.common import utils
 from neutron.conf.db import l3_hamode_db
 from neutron.db import _utils as db_utils
 from neutron.db.availability_zone import router as router_az_db
@@ -143,7 +143,10 @@ class L3_HA_NAT_db_mixin(l3_dvr_db.L3_NAT_with_dvr_db_mixin,
                         context, network_id=network_id, vr_id=vr_id)
                     allocation.create()
 
-                    router_db.extra_attributes.ha_vr_id = allocation.vr_id
+                    router_db.extra_attributes.update(
+                        {'ha_vr_id': allocation.vr_id})
+                    context.session.add(router_db.extra_attributes)
+
                     LOG.debug(
                         "Router %(router_id)s has been allocated a ha_vr_id "
                         "%(ha_vr_id)d.",
@@ -200,7 +203,7 @@ class L3_HA_NAT_db_mixin(l3_dvr_db.L3_NAT_with_dvr_db_mixin,
             # and the process is started over where the existing
             # network will be selected.
             raise db_exc.DBDuplicateEntry(columns=['tenant_id'])
-        return ha_network
+        return None, ha_network
 
     def _add_ha_network_settings(self, network):
         if cfg.CONF.l3_ha_network_type:
@@ -267,7 +270,7 @@ class L3_HA_NAT_db_mixin(l3_dvr_db.L3_NAT_with_dvr_db_mixin,
                 context, port_id=port_id, router_id=router_id)
             portbinding.create()
 
-            return portbinding
+            return None, portbinding
         except db_exc.DBReferenceError as e:
             with excutils.save_and_reraise_exception() as ctxt:
                 if isinstance(e.inner_exception, sql_exc.IntegrityError):
@@ -285,7 +288,7 @@ class L3_HA_NAT_db_mixin(l3_dvr_db.L3_NAT_with_dvr_db_mixin,
         # the exception will mangle the state, cause the delete call to fail,
         # and end up relying on the DB rollback to remove the port instead of
         # proper delete_port call.
-        if context.session.is_active:
+        if utils.is_session_active(context.session):
             raise RuntimeError(_('add_ha_port cannot be called inside of a '
                                  'transaction.'))
         args = {'tenant_id': '',
@@ -349,25 +352,15 @@ class L3_HA_NAT_db_mixin(l3_dvr_db.L3_NAT_with_dvr_db_mixin,
         return super(L3_HA_NAT_db_mixin,
                      self)._get_device_owner(context, router)
 
-    @n_utils.transaction_guard
-    def _ensure_vr_id_and_network(self, context, router_db):
-        """Attach vr_id to router while tolerating network deletes."""
-        creator = functools.partial(self._ensure_vr_id,
-                                    context, router_db)
-        dep_getter = functools.partial(self.get_ha_network,
-                                       context, router_db.tenant_id)
-        dep_creator = functools.partial(self._create_ha_network,
-                                        context, router_db.tenant_id)
-        dep_deleter = functools.partial(self._delete_ha_network, context)
-        dep_id_attr = 'network_id'
-        return n_utils.create_object_with_dependency(
-            creator, dep_getter, dep_creator, dep_id_attr, dep_deleter)[1]
-
     @registry.receives(resources.ROUTER, [events.BEFORE_CREATE],
                        priority_group.PRIORITY_ROUTER_EXTENDED_ATTRIBUTE)
+    def _before_router_create_handler(self, resource, event, trigger,
+                                      payload):
+        return self._before_router_create(event, payload.context,
+                                          payload.latest_state)
+
     @db_api.retry_if_session_inactive()
-    def _before_router_create(self, resource, event, trigger,
-                              context, router, **kwargs):
+    def _before_router_create(self, event, context, router):
         """Event handler to create HA resources before router creation."""
         if not self._is_ha(router):
             return
@@ -379,9 +372,11 @@ class L3_HA_NAT_db_mixin(l3_dvr_db.L3_NAT_with_dvr_db_mixin,
 
     @registry.receives(resources.ROUTER, [events.PRECOMMIT_CREATE],
                        priority_group.PRIORITY_ROUTER_EXTENDED_ATTRIBUTE)
-    def _precommit_router_create(self, resource, event, trigger, context,
-                                 router, router_db, **kwargs):
+    def _precommit_router_create(self, resource, event, trigger, payload):
         """Event handler to set ha flag and status on creation."""
+        context = payload.context
+        router = payload.latest_state
+        router_db = payload.metadata['router_db']
         is_ha = self._is_ha(router)
         router['ha'] = is_ha
         self.set_extra_attr_value(context, router_db, 'ha', is_ha)
@@ -399,8 +394,11 @@ class L3_HA_NAT_db_mixin(l3_dvr_db.L3_NAT_with_dvr_db_mixin,
 
     @registry.receives(resources.ROUTER, [events.AFTER_CREATE],
                        priority_group.PRIORITY_ROUTER_EXTENDED_ATTRIBUTE)
-    def _after_router_create(self, resource, event, trigger, context,
-                             router_id, router, router_db, **kwargs):
+    def _after_router_create(self, resource, event, trigger, payload):
+        context = payload.context
+        router_id = payload.resource_id
+        router = payload.latest_state
+        router_db = payload.metadata['router_db']
         if not router['ha']:
             return
         try:
@@ -466,10 +464,14 @@ class L3_HA_NAT_db_mixin(l3_dvr_db.L3_NAT_with_dvr_db_mixin,
 
     @registry.receives(resources.ROUTER, [events.AFTER_UPDATE],
                        priority_group.PRIORITY_ROUTER_EXTENDED_ATTRIBUTE)
-    def _reconfigure_ha_resources(self, resource, event, trigger, context,
-                                  router_id, old_router, router, router_db,
-                                  **kwargs):
+    def _reconfigure_ha_resources(self, resource, event, trigger, payload):
         """Event handler to react to changes after HA flag has been updated."""
+        context = payload.context
+        router_id = payload.resource_id
+        old_router = payload.states[0] if (len(payload.states) > 1) else None
+        router = payload.latest_state
+        router_db = payload.metadata['router_db']
+
         ha_changed = old_router['ha'] != router['ha']
         if not ha_changed:
             return
@@ -522,9 +524,10 @@ class L3_HA_NAT_db_mixin(l3_dvr_db.L3_NAT_with_dvr_db_mixin,
 
     @registry.receives(resources.ROUTER, [events.PRECOMMIT_DELETE],
                        priority_group.PRIORITY_ROUTER_EXTENDED_ATTRIBUTE)
-    def _release_router_vr_id(self, resource, event, trigger, context,
-                              router_db, **kwargs):
+    def _release_router_vr_id(self, resource, event, trigger, payload):
         """Event handler for removal of VRID during router delete."""
+        context = payload.context
+        router_db = payload.latest_state
         if router_db.extra_attributes.ha:
             ha_network = self.get_ha_network(context,
                                              router_db.tenant_id)
@@ -534,9 +537,16 @@ class L3_HA_NAT_db_mixin(l3_dvr_db.L3_NAT_with_dvr_db_mixin,
 
     @registry.receives(resources.ROUTER, [events.AFTER_DELETE],
                        priority_group.PRIORITY_ROUTER_EXTENDED_ATTRIBUTE)
+    def _cleanup_ha_network_handler(self, resource, event, trigger, payload):
+
+        return self._cleanup_ha_network(event,
+                                        payload.context,
+                                        payload.resource_id,
+                                        payload.latest_state)
+
     @db_api.retry_if_session_inactive()
-    def _cleanup_ha_network(self, resource, event, trigger, context,
-                            router_id, original, **kwargs):
+    def _cleanup_ha_network(self, event, context,
+                            router_id, original):
         """Event handler to attempt HA network deletion after router delete."""
         if not original['ha']:
             return
@@ -551,6 +561,7 @@ class L3_HA_NAT_db_mixin(l3_dvr_db.L3_NAT_with_dvr_db_mixin,
         for agent in self.get_l3_agents_hosting_routers(context, [router_id]):
             self.remove_router_from_l3_agent(context, agent['id'], router_id)
 
+    @db_api.CONTEXT_READER
     def get_ha_router_port_bindings(self, context, router_ids, host=None):
         if not router_ids:
             return []
@@ -686,8 +697,8 @@ class L3_HA_NAT_db_mixin(l3_dvr_db.L3_NAT_with_dvr_db_mixin,
     def _set_router_states(cls, context, bindings, states):
         for binding in bindings:
             try:
-                # NOTE(ralonsoh): to be migrated to the new facade.
-                with context.session.begin(subtransactions=True):
+                with db_api.CONTEXT_WRITER.using(context):
+                    context.session.add(binding)
                     binding.state = states[binding.router_id]
             except (orm.exc.StaleDataError, orm.exc.ObjectDeletedError):
                 # Take concurrently deleted routers in to account
@@ -697,8 +708,9 @@ class L3_HA_NAT_db_mixin(l3_dvr_db.L3_NAT_with_dvr_db_mixin,
     def update_routers_states(self, context, states, host):
         """Receive dict of router ID to state and update them all."""
 
-        bindings = self.get_ha_router_port_bindings(
-            context, router_ids=states.keys(), host=host)
+        with db_api.CONTEXT_READER.using(context):
+            bindings = self.get_ha_router_port_bindings(
+                context, router_ids=states.keys(), host=host)
         self._set_router_states(context, bindings, states)
         self._update_router_port_bindings(context, states, host)
 
