@@ -53,7 +53,7 @@ class TestSbApi(BaseOvnIdlTest):
                 {'external_ids': {'ovn-bridge-mappings':
                                   'public:br-ex,private:br-0'}},
                 {'external_ids': {'ovn-bridge-mappings':
-                                  'public:br-ex,public2:br-ex'}},
+                                  'public:br-ex,public2:br-ex2'}},
                 {'external_ids': {'ovn-bridge-mappings':
                                   'public:br-ex'}},
             ]
@@ -81,13 +81,6 @@ class TestSbApi(BaseOvnIdlTest):
         our_chassis = {c['name'] for c in self.data['chassis']}
         self.assertLessEqual(our_chassis, chassis_list)
 
-    def test_get_chassis_data_for_ml2_bind_port(self):
-        host = self.data['chassis'][0]['hostname']
-        dp, iface, phys = self.api.get_chassis_data_for_ml2_bind_port(host)
-        self.assertEqual('', dp)
-        self.assertEqual('', iface)
-        self.assertItemsEqual(phys, ['private', 'public'])
-
     def test_chassis_exists(self):
         self.assertTrue(self.api.chassis_exists(
             self.data['chassis'][0]['hostname']))
@@ -99,7 +92,17 @@ class TestSbApi(BaseOvnIdlTest):
         self.assertGreaterEqual(set(mapping.keys()),
                                 {c['name'] for c in self.data['chassis']})
 
-    def _add_switch_port(self, chassis_name, type='localport'):
+    def test_multiple_physnets_in_one_bridge(self):
+        self.data = {
+            'chassis': [
+                {'external_ids': {'ovn-bridge-mappings': 'p1:br-ex,p2:br-ex'}}
+            ]
+        }
+        self.load_test_data()
+        self.assertRaises(ValueError, self.api.get_chassis_and_physnets)
+
+    def _add_switch_port(self, chassis_name,
+                         type=ovn_const.LSP_TYPE_LOCALPORT):
         sname, pname = (utils.get_rand_device_name(prefix=p)
                         for p in ('switch', 'port'))
         chassis = self.api.lookup('Chassis', chassis_name)
@@ -123,14 +126,7 @@ class TestSbApi(BaseOvnIdlTest):
         val = str(uuid.uuid4())
         self.assertIsNone(self.api.get_metadata_port_network(val))
 
-    def test_set_get_chassis_metadata_networks(self):
-        name = self.data['chassis'][0]['name']
-        nets = [str(uuid.uuid4()) for _ in range(3)]
-        self.api.set_chassis_metadata_networks(name, nets).execute(
-            check_error=True)
-        self.assertEqual(nets, self.api.get_chassis_metadata_networks(name))
-
-    def test_get_network_port_bindings_by_ip(self):
+    def _create_bound_port_with_ip(self):
         chassis, switch, port, binding = self._add_switch_port(
             self.data['chassis'][0]['name'])
         mac = 'de:ad:be:ef:4d:ad'
@@ -143,10 +139,28 @@ class TestSbApi(BaseOvnIdlTest):
             port.name, [mac_ip]).execute(check_error=True)
         self.assertTrue(pb_update_event.wait())
         self.api.lsp_bind(port.name, chassis.name).execute(check_error=True)
+
+        return binding, ipaddr, switch
+
+    def test_get_network_port_bindings_by_ip(self):
+        binding, ipaddr, _ = self._create_bound_port_with_ip()
         result = self.api.get_network_port_bindings_by_ip(
             str(binding.datapath.uuid), ipaddr)
         self.assertIn(binding, result)
 
+    def test_get_network_port_bindings_by_ip_with_unbound_port(self):
+        binding, ipaddr, switch = self._create_bound_port_with_ip()
+        unbound_port_name = utils.get_rand_device_name(prefix="port")
+        mac_ip = "de:ad:be:ef:4d:ab %s" % ipaddr
+        with self.nbapi.transaction(check_error=True) as txn:
+            txn.add(
+                self.nbapi.lsp_add(switch.name, unbound_port_name, type=type))
+            txn.add(self.nbapi.lsp_set_addresses(unbound_port_name, [mac_ip]))
+        result = self.api.get_network_port_bindings_by_ip(
+            str(binding.datapath.uuid), ipaddr)
+        self.assertIn(binding, result)
+        self.assertEqual(1, len(result))
+
     def test_get_ports_on_chassis(self):
         chassis, switch, port, binding = self._add_switch_port(
             self.data['chassis'][0]['name'])
@@ -154,14 +168,6 @@ class TestSbApi(BaseOvnIdlTest):
         self.assertEqual([binding],
                          self.api.get_ports_on_chassis(chassis.name))
 
-    def test_get_logical_port_chassis_and_datapath(self):
-        chassis, switch, port, binding = self._add_switch_port(
-            self.data['chassis'][0]['name'])
-        self.api.lsp_bind(port.name, chassis.name).execute(check_error=True)
-        self.assertEqual(
-            (chassis.name, str(binding.datapath.uuid)),
-            self.api.get_logical_port_chassis_and_datapath(port.name))
-
 
 class TestNbApi(BaseOvnIdlTest):
 
@@ -226,7 +232,7 @@ class TestNbApi(BaseOvnIdlTest):
         exp_values = [(lb['name'], lb['external_ids'])
                       for lb in self.data['lbs']]
         lbs_values = [(lb.name, lb.external_ids) for lb in lbs]
-        self.assertItemsEqual(exp_values, lbs_values)
+        self.assertCountEqual(exp_values, lbs_values)
 
     def test_get_router_floatingip_lbs(self):
         f = self.nbapi.get_router_floatingip_lbs
@@ -242,7 +248,7 @@ class TestNbApi(BaseOvnIdlTest):
             lbs_values = [(lb.name, lb.external_ids)
                           for lb in f(exp_router_name)]
             self.assertTrue(exp_values)
-            self.assertItemsEqual(exp_values, lbs_values)
+            self.assertCountEqual(exp_values, lbs_values)
 
     def test_get_floatingip_in_nat_or_lb(self):
         f = self.nbapi.get_floatingip_in_nat_or_lb
